/* actual parsing

   enumeration approaches - generate all possible parses/sentences and
   see if they match the given sentence!  what could be slow about
   that?
 
   basic: traverse grammar

   since some grammars can generate infinite sentences, use depth
   pruning: max depth = length of sentence
   
   possible improvement:
      sentence pruning: test generated terminals against sentence as we go
*/

#include <stdlib.h>  /* malloc */
#include "utils.h"
#include "parse.h"


/* unlike for CYK, we want the rules that generate nonterminals to be
   mixed into the grammar alongside the rules that generate two
   nonterminals.  we'll use a nonterminal_2 = TERMINAL to indicate
   that a rule generates the terminal specified in nonterminal_1.
   this is kinda klugy but at least we avoid having to define a new
   grammar type... */


#define TERMINAL -1


/* frees the terminal rule after merging new nonterminal version*/
void add_terminal_rules(terminal_lhs *r, int t, cfg *g)
{
  if (r) {
    int nt = r->nonterminal;
    terminal_lhs *next = r->next;
    nonterminal_rhs *newr = (nonterminal_rhs *) malloc(sizeof(nonterminal_rhs));
    
    if (! newr) panic("failed to alloc klugy nonterminal_rhs");
    newr->nonterminal_1 = t;
    newr->nonterminal_2 = TERMINAL;
    newr->probability = r->probability;
    /* push new rule onto list */
    newr->next = g->nonterminal_rules[nt];
    g->nonterminal_rules[nt] = newr;
    free(r);
    add_terminal_rules(next, t, g);
  }
}


/* modifies the given grammar */
void merge_terminal_rules(cfg *g)
{
  int size = symtab_count(g->st);
  int t;

  for (t = 0; t < size; t++) {
    add_terminal_rules(g->terminal_rules[t], t, g);
    g->terminal_rules[t] = NULL;
  }
}


/************************************************************
 generating all parses

 each NT encountered as we generate a sentence is a decision point
 where multiple options may be possible - ie, multiple rules to
 generate from that nt.  to enumerate all possibilities, keep a stack
 of "counters", one per decision point, keeping track of which option
 should be considered for each decision.  these can be interpreted as
 a set of digits like an odometer, most recent on the right and oldest
 on the left.  a particular digit position in the counter might have
 only one possible digit value or many, depending on the number of
 rules for the NT.  a counter is allocated if a decision point is
 reached and no counter exists, and deallocated once a counter uses up
 all its options.  during extraction, the option specified by the
 "counter" is taken at each decision point.  after each extraction,
 the rightmost digit is incremented, where "carrying" involves
 deallocation.

 protocol:
    at each decision:
      options_needed, if so, set_options
      current_option
    after each parse extraction:
      increment counters

 ************************************************************/


#define INITIAL_OA_SIZE 4


typedef struct option_array {
  int size;
  int count;
  /* array of length size currently holding count linked lists of
     remaining options.  in each list, the current option is first */
  nonterminal_rhs **options;
  int current_index;  /* which decision we have just processed */
} option_array;



static option_array *make_option_array(void)
{
  option_array *oa = (option_array *) malloc(sizeof(option_array));
  nonterminal_rhs **options;

  if (! oa) panic("failed to malloc option_array");
  options = (nonterminal_rhs **) malloc(sizeof(nonterminal_rhs *) * INITIAL_OA_SIZE);
  if (! options) panic("failed to malloc options");
  oa->size = INITIAL_OA_SIZE;
  oa->count = 0;
  oa->options = options;
  oa->current_index = -1;
  return oa;
}


static void free_option_array(option_array *oa)
{
  /* options themselves are part of grammar, which will be freed separately */
  free(oa->options);
  free(oa);
}


/* increments current decision index.  returns non-zero iff no entry
   exists and options must be supplied */
static bool options_needed(option_array *oa)
{
  oa->current_index++;
  return (oa->current_index == oa->count);
}


/* load options for current decision into oa */
static void set_options(option_array *oa, nonterminal_rhs *options)
{
  int i = oa->current_index;
  
  if (i != oa->count) panic("why was set_options called?");
  if (i == oa->size) {
    int newsize = oa->size * 2;
    nonterminal_rhs **new = (nonterminal_rhs **) realloc(oa->options,
							 sizeof(nonterminal_rhs *) * newsize);
    
    if (! new) panic("failed to realloc options");
    oa->size = newsize;
    oa->options = new;
  }
  if (!options) panic("no choices when extracting multiple parses?");
  oa->options[i] = options;
  oa->count++;
}


/* called at each decision point during extraction */
static nonterminal_rhs *current_option(option_array *oa)
{
  return oa->options[oa->current_index];
}


/* returns whether there exist further parses */
static bool increment_last(option_array *oa)
{
  int last = oa->count - 1;

  if (last >= 0) {
    nonterminal_rhs *remaining = oa->options[last];
    nonterminal_rhs *next = remaining->next;
    
    /* we are not freeing options, just stepping through the list */
    if (next) {
      oa->options[last] = next;
      return TRUE;
    } else {
      /* used up, time to "carry" */
      oa->count--;
      return increment_last(oa);
    }
  } else {
    return FALSE;
  }
}

		     
/* called between extractions.  deallocates counters if necssary.
   resets current_index.  returns whether to keep going. */
static bool increment_counters(option_array *oa)
{
  if (oa->current_index != (oa->count - 1)) panic("didn't finish extracting?");
  if (increment_last(oa)) {
    oa->current_index = -1;
    return TRUE;
  } else {
    return FALSE;
  }
}


parse *generate_terminal(int t)
{
  parse *p = (parse *) malloc(sizeof(parse));
  
  if (! p) panic("failed to malloc terminal parse");
  p->symbol = t;
  p->first = NULL;
  return p;
}


/* generate phrase structure rooted at an nt using g.  follow decision
   counters stored in oa.  prune using depth_remaining.  improvement:
   prune using s.  returns NULL if unable to generate full phrase.  */
parse *generate_parse(cfg *g, int nt, option_array *oa, int depth_remaining, token_seq *s)
{
  if (depth_remaining < 0) {
    return NULL;
  } else {
    nonterminal_rhs *r;
    
    if (options_needed(oa)) set_options(oa, g->nonterminal_rules[nt]);
    r = current_option(oa);
    if (! r) panic("no way of generating one of the non-terminals!");
    if (r->nonterminal_2 == TERMINAL) {
      parse *p = (parse *) malloc(sizeof(parse));
      
      if (! p) panic("can't malloc nt->t parse while generating");
      p->symbol = nt;
      p->first = generate_terminal(r->nonterminal_1);
      p->second = NULL;
      p->probability = r->probability;
      return p;
    } else {
      parse *first = generate_parse(g, r->nonterminal_1, oa, depth_remaining-1, s);
      parse *second = generate_parse(g, r->nonterminal_2, oa, depth_remaining-1, s);

      if (first && second) {
	parse *p = (parse *) malloc(sizeof(parse));
      
	if (! p) panic("can't malloc nt-> nt1 nt2 parse while generating");
	p->symbol = nt;
	p->first = first;
	p->second = second;
	p->probability = first->probability * second->probability * r->probability;
	return p;
      } else {
	if (first) free_parse(first);
	if (second) free_parse(second);
	return NULL;
      }
    }
  }
}


bool matches_at(parse *p, token_seq *s, int *ip)
{
  if (*ip < s->len) {
    if (p->first == NULL) {
      /* terminal */
      if (p->symbol == s->tokens[*ip]) {
	(*ip)++;
	return TRUE;
      } else
	return FALSE;
    } else if (p->second == NULL)
      /* nt covering t */
      return matches_at(p->first, s, ip);
    else
      return (matches_at(p->first, s, ip) &&
	      matches_at(p->second, s, ip));
  } else
    return FALSE;
}


bool matches_sentence(parse *p, token_seq *s)
{
  int i = 0;

  return matches_at(p, s, &i);
}


parses *enum_all_parses(token_seq *s, cfg *g)
{
  int start = symtab_id_for_name(g->st, START_NONTERMINAL_NAME);
  option_array *oa = make_option_array();
  parses *ps = NULL;
  
  merge_terminal_rules(g);
  do {
    parse *p = generate_parse(g, start, oa, s->len, s);
    
    if (p) {
      if (matches_sentence(p, s)) {
	parses *new = (parses *) malloc(sizeof(parses));
	if (! new) panic("failed to malloc another parses");
	new->parse = p;
	new->next = ps;
	ps = new;
      } else {
	free_parse(p);
      }
    }
  } while (increment_counters(oa));
  free_option_array(oa);
  return ps;
}


/************************************************************
 * most probable parse
 ************************************************************/


parses *enum_most_probable_parse(token_seq *s, cfg *g)
{
  /* the lazy approach, in keeping with the theme of enumeration */
  parses *ps = enum_all_parses(s, g);

  /* NOW SELECT JUST THE MOST PROBABLE */
  panic("enum_most_probable_parse not implemented");
  return ps;
}


parses *enum_parse(token_seq *s, cfg *g)
{
  return enum_all_parses(s, g);
}


/* EOF */
