/* UNH CS 758 assignment 7

*/

#include <stdio.h>
#include <stdlib.h>  /* free, EXIT_SUCCESS, rand */
#include <string.h>  /* strtok */
#include <ctype.h>   /* islower, isspace */
#include "utils.h"
#include "parse.h"


#define INITIAL_G_SIZE 4
/* for grammar file */
#define MAX_LINE_LEN 200
#define WHITESPACE " \a\b\f\n\r\t\v"

/* for sentences */
#define INITIAL_S_SIZE 4
#define MAX_S_LEN 512


/************************************************************
 * making and freeing grammars
 ************************************************************/


cfg *make_grammar(void)
{
  cfg *g = (cfg *) malloc(sizeof(cfg));
  nonterminal_rhs **nts = (nonterminal_rhs **) calloc(INITIAL_G_SIZE,
						      sizeof(nonterminal_rhs *));
  terminal_lhs **ts = (terminal_lhs **) calloc(INITIAL_G_SIZE,
						      sizeof(terminal_lhs *));
  
  if (!g) panic("can't malloc grammar");
  if (!nts) panic("can't malloc nonterminal rule array");
  if (!ts) panic("can't malloc terminal rule array");
  g->st = make_symtab();
  g->size = INITIAL_G_SIZE;
  g->nonterminal_rules = nts;
  g->terminal_rules = ts;
  return g;
}


void ensure_grammar_space(cfg *g)
{
  int oldsize = g->size;
  
  if ((symtab_count(g->st) + 1) > oldsize) {
    int newsize = oldsize * 2;
    
    g->nonterminal_rules = (nonterminal_rhs **) recalloc(g->nonterminal_rules, oldsize, newsize,
							 sizeof(nonterminal_rhs *));
    g->terminal_rules = (terminal_lhs **) recalloc(g->terminal_rules, oldsize, newsize,
						   sizeof(terminal_lhs *));
    g->size = newsize;
  }
}


void free_nt_rhss(nonterminal_rhs *r)
{
  if (r) {
    nonterminal_rhs *next = r->next;

    free(r);
    free_nt_rhss(next);
  }
}


void free_t_lhss(terminal_lhs *r)
{
  if (r) {
    terminal_lhs *next = r->next;

    free(r);
    free_t_lhss(next);
  }
}


void free_grammar(cfg *g)
{
  int size = symtab_count(g->st);
  int i;

  for (i = 0; i < size; i++) {
    free_nt_rhss(g->nonterminal_rules[i]);
  }
  free(g->nonterminal_rules);
  for (i = 0; i < size; i++) {
    free_t_lhss(g->terminal_rules[i]);
  }
  free(g->terminal_rules);
  free_symtab(g->st);
  free(g);
}


/************************************************************
 * read grammar
 ************************************************************/


/* skips over whitespace characters in infile.  returns EOF iff EOF */
int skip_white(FILE *infile)
{
  int c;
  
  do {
    c = fgetc(infile);
  } while (isspace(c));
  if (c != EOF) {
    ungetc(c, infile);
  }
  return c;
}


/* read through next \n or EOF, return EOF iff EOF */
int skip_line(FILE *infile)
{
  int c;
  
  do {
    c = fgetc(infile);
  } while ((c != EOF) && (c != '\n'));
  return c;
}


/* returns EOF iff EOF or error */
int skip_noncontent_lines(FILE *infile)
{  
  if (skip_white(infile) == EOF)
    return EOF;
  else {
    int c = fgetc(infile);
    
    if (c == '#') {
      if (skip_line(infile) == EOF)
	return EOF;
      else
	return skip_noncontent_lines(infile);
    }
    return ungetc(c, infile);
  }
}


/* returns pointer to static buffer holding next line of input, or
   NULL iff EOF or error.  lines starting with # or whitespace are skipped */
char *read_line(FILE *infile)
{
  static char buf[MAX_LINE_LEN+1];
  
  if (skip_noncontent_lines(infile) == EOF)
    return NULL;
  return fgets(buf, MAX_LINE_LEN, infile);
}


double get_prob()
{
  char *tok = strtok(NULL, WHITESPACE);

  if (! tok) panic("didn't find probability at end of grammar rule");
  return atof(tok);
}


/* returns 0 on EOF */
int read_rule(FILE *infile, cfg *g)
{
  char *l = read_line(infile);

  if (l) {
    char *nonterm = strtok(l, WHITESPACE);
    int nt;
    char *tok;

    ensure_grammar_space(g);
    if (! nonterm) panic("didn't find first token in grammar rule");
    nt = symtab_id_for_name(g->st, nonterm);
    tok = strtok(NULL, WHITESPACE);
    if (! tok) panic("didn't find second token in grammar rule");
    if (islower(*tok)) {
      /* terminal rule */
      int t = symtab_id_for_name(g->st, tok);
      terminal_lhs *r = (terminal_lhs *) malloc(sizeof(terminal_lhs));

      if (!r) panic("can't malloc new terminal rule!");
      r->nonterminal = nt;
      r->probability = get_prob();
      r->next = g->terminal_rules[t];
      g->terminal_rules[t] = r;
    } else {
      /* nonterminal rule */
      int nt1 = symtab_id_for_name(g->st, tok);
      int nt2;
      nonterminal_rhs *r;

      tok = strtok(NULL, WHITESPACE);
      if (! tok) panic("didn't find second RHS nonterminal in grammar rule");
      nt2 = symtab_id_for_name(g->st, tok);
      r = (nonterminal_rhs *) malloc(sizeof(nonterminal_rhs));
      if (!r) panic("can't malloc new nonterminal rule!");
      r->nonterminal_1 = nt1;
      r->nonterminal_2 = nt2;
      r->probability = get_prob();      
      r->next = g->nonterminal_rules[nt];
      g->nonterminal_rules[nt] = r;
    }
    return 1;
  } else
    return 0;
}


cfg *read_grammar(FILE *infile)
{
  cfg *g = make_grammar();

  while (read_rule(infile, g)) {
  }
  return g;
}


/************************************************************
 * print grammar
 ************************************************************/


void fprint_nt_rhss(FILE *outfile, cfg *g, nonterminal_rhs *r)
{
  if (r) {
    fprintf(outfile, "   -> %s %s %6f\n",
	    symtab_name_for_id(g->st, r->nonterminal_1),
	    symtab_name_for_id(g->st, r->nonterminal_2),
	    r->probability);
    fprint_nt_rhss(outfile, g, r->next);
  }
}


void fprint_t_lhss(FILE *outfile, cfg *g, terminal_lhs *r)
{
  if (r) {
    fprintf(outfile, "   <- %s %6f\n",
	    symtab_name_for_id(g->st, r->nonterminal),
	    r->probability);
    fprint_t_lhss(outfile, g, r->next);
  }
}


void fprint_grammar(FILE *outfile, cfg *g)
{
  int i;
  
  fprintf(outfile, "-- CFG with allocated size %d --\n", g->size);
  fprint_symtab(outfile, g->st);
  for (i = 0; i < g->size; i++) {
    if (g->nonterminal_rules[i]) {
      fprintf(outfile, "%s ->\n", symtab_name_for_id(g->st, i));
      fprint_nt_rhss(outfile, g, g->nonterminal_rules[i]);
    }
  }
  for (i = 0; i < g->size; i++) {
    if (g->terminal_rules[i]) {
      fprintf(outfile, "%s <-\n", symtab_name_for_id(g->st, i));
      fprint_t_lhss(outfile, g, g->terminal_rules[i]);
    }
  }
  fprintf(outfile, "-- end CFG --\n");
}


/************************************************************
 * the sentence
 ************************************************************/


token_seq *read_sentence(FILE *infile, cfg *g)
{
  char buf[MAX_S_LEN];
  token_seq *s = (token_seq *) malloc(sizeof(token_seq));
  int *toks = (int *) malloc(sizeof(int) * INITIAL_S_SIZE);
  char *tok;

  if (!s) panic("can't malloc sentence");
  if (!toks) panic("can't malloc sentence toks");
  s->size = INITIAL_S_SIZE;
  s->len = 0;
  s->tokens = toks;
  if (! fgets(buf, MAX_S_LEN, infile)) panic("EOF or error reading sentence");
  tok = strtok(buf, WHITESPACE);
  while (tok != NULL) {
    int symbol = symtab_id_for_name(g->st, tok);
    
    if ((s->len + 1) >= s->size) {
      s->size = s->size * 2;
      s->tokens = (int *) realloc(s->tokens, sizeof(int *) * s->size);
      if (! s->tokens) panic("failed to realloc sentence");
    }
    s->tokens[s->len++] = symbol;
    tok = strtok(NULL, WHITESPACE);
  }
  return s;
}


void fprint_sentence(FILE *outfile, token_seq *s, cfg *g)
{
  int i;

  fprintf(outfile, "sentence (%d/%d):", s->len, s->size);
  for (i = 0; i < s->len; i++) {
    fprintf(outfile, " %s", symtab_name_for_id(g->st, s->tokens[i]));
  }
  fprintf(outfile, "\n");
}


void free_sentence(token_seq *s)
{
  free(s->tokens);
  free(s);
}


/************************************************************
 * parses
 ************************************************************/


/* recursive helper for fprint_parse */
void fpf_parse(FILE *outfile, parse *p, cfg *g)
{
  if (p->first == NULL) {
    /* terminal */
    fprintf(outfile, "%s", symtab_name_for_id(g->st, p->symbol));
  } else {
    /* either nonterminal covering a terminal or binary branching
       nonterminal.  either way, print first. */
    fprintf(outfile, "%s(", symtab_name_for_id(g->st, p->symbol));
    fpf_parse(outfile, p->first, g);
    if (p->second != NULL) {
      /* branching, so print second */
      fprintf(outfile, " ");
      fpf_parse(outfile, p->second, g);
    }
    fprintf(outfile, ")");
  }
}


void fprint_parse(FILE *outfile, parse *p, cfg *g)
{
  if (p == NULL)
    fprintf(outfile, "No parse found!\n");
  else {
    fpf_parse(outfile, p, g);
    fprintf(outfile, " %6f\n", p->probability);
  }
}


void fprint_parses(FILE *outfile, parses *ps, cfg *g)
{
  if (ps) {
    fprint_parse(outfile, ps->parse, g);
    fprint_parses(outfile, ps->next, g);
  }
}


/* makes a list of a single parse */
parses *wrap_parse(parse *p)
{
  parses *ps = (parses *) malloc(sizeof(parses));

  if (! ps) panic("couldn't malloc single parses");
  ps->parse = p;
  ps->next = NULL;
  return ps;
}


void free_parse(parse *p)
{
  if (p->first) {
    free_parse(p->first);
    if (p->second) {
      free_parse(p->second);
    }
  }
  free(p);
}


void free_parses(parses *ps)
{
  if (ps) {
    parses *next = ps->next;

    free_parse(ps->parse);
    free(ps);
    free_parses(next);
  }
}


/* EOF */
