/* symbol table

   implemented as a hash table of char* -> int and an array[int] -> char*
   hash table uses collision resolution by linear probing
*/

#include <stdio.h>
#include <stdlib.h>  /* malloc */
#include <string.h>  /* strcmp */
#include "utils.h"
#include "symbol_table.h"


/* table likely small, so a low max load factor like 0.25 is ok */
#define INVERSE_MAX_LOAD_FACTOR 3
/* should be set to something like 16 after testing */
#define INITIAL_SIZE 4


/******** allocating / deallocating *********/


symbol_table *make_symtab(void)
{
  symbol_table *st = (symbol_table *) malloc(sizeof(symbol_table));
  st_entry *entries;
  char **keys;

  if (! st) panic("failed to malloc symbol table");
  entries = (st_entry *) calloc(INITIAL_SIZE, sizeof(st_entry));
  if (! entries) panic("failed to calloc initial symtab entries");
  keys = (char **) malloc(sizeof(char *) * INITIAL_SIZE);
  if (! keys) panic("failed to malloc initial symtab keys");
  st->size = INITIAL_SIZE;
  st->entries = entries;
  st->count = 0;
  st->keys = keys;
  return st;
}


/* also frees the strings! */
void free_symtab(symbol_table *st)
{
  int i;
  
  for (i = 0; i < st->count; i++) {
    free(st->keys[i]);
  };
  free(st->entries);
  free(st->keys);
  free(st);
}


/******** I/O *********/


void fprint_symtab(FILE *outfile, symbol_table *st)
{
  int size = st->size;
  st_entry *es = st->entries;
  char **ks = st->keys;
  int i;

  fprintf(outfile, "-- symbol table of size %d with %d elements --\n", size, st->count);
  for (i = 0; i < size; i++) {
    if (es[i].key) {
      fprintf(outfile, "'%s' -> %d\n", es[i].key, es[i].value);
    }
  }
  for (i = 0; i < st->count; i++) {
    fprintf(outfile, "%d -> '%s'\n", i, ks[i]);
  }
  fprintf(outfile, "-- end of symbol table --\n");
}


/***** utilities *****/


unsigned int hash (char *key)
{
  unsigned int hash = 0;
  int i;

  for (i = 0; key[i] != '\0'; i++)
    hash = (hash * 63) + key[i];
  return hash;
}


/* returns previously assigned id value for the given key, or a
   negative number iff key not found */
int find (symbol_table *st, char *key)
{
  int start = (hash(key) % st->size);
  int i;
  int found = FALSE;

  for (i = start; st->entries[i].key != NULL; ) {
    if (strcmp(key,st->entries[i].key) == 0) {
      found = TRUE;
      break;
    }
    i = (i + 1) % st->size;
    if (i == start) panic("symtab_lookup wrapped around: is table completely full?");
  }
  if (found)
    return st->entries[i].value;
  else
    return -1;
}


/* returns empty slot for key's entry.  does NOT check if key is
   already present */
int empty_slot (char *key, int size, st_entry *entries)
{
  int start = (hash(key) % size);
  int i;
  
  for (i = start; entries[i].key != NULL; ) {
    i = (i + 1) % size;
    if (i == start) panic("symtab_lookup wrapped around: is table completely full?");
  }
  return i;
}


st_entry *realloc_entries (symbol_table *st, int newsize)
{
  int oldsize = st->size;
  st_entry *old = st->entries;
  st_entry *new = (st_entry *) calloc(newsize, sizeof(st_entry));
  int i;
  
  if (!new) panic("failed to calloc new symbol table entries");
  for (i = 0; i < oldsize; i++) {
    char *key = old[i].key;
    
    if (key) {
      int j = empty_slot(key, newsize, new);
      new[j].key = key;
      new[j].value = old[i].value;
    }
  }
  free(old);
  return new;
}


char **realloc_keys (symbol_table *st, int newsize)
{
  char **new = (char **) realloc(st->keys, sizeof(char *) * newsize);
  if (!new) panic("failed to realloc new symbol table keys");
  return new;
}
    

/* assumes count has already been incremented to new desired value */
void ensure_load_factor (symbol_table *st)
{
  int size = st->size;
  int min = st->count * INVERSE_MAX_LOAD_FACTOR;

  if (size < min) {
    int newsize = size * 2;
    
    st->entries = realloc_entries(st, newsize);
    st->keys = realloc_keys(st, newsize);    
    st->size = newsize;
  }
}


/***** main interface functions *****/


/* number of mappings currently stored.  count-1 is maximum symbol id
   currently in use.  */
int symtab_count(symbol_table *st)
{
  return st->count;
}


/* returns unique int corresponding to name.  if name has not been
   seen before, stash a copy of it and associate it with a new int. */
int symtab_id_for_name(symbol_table *st, char *name)
{
  int value = find(st, name);

  if (value < 0) {
    /* make new */
    char *key = strcpy_new(name);
    st_entry *e;
    int i;
    
    value = st->count++;
    ensure_load_factor(st);
    st->keys[value] = key;
    i = empty_slot(key, st->size, st->entries);
    e = st->entries + i;
    e->key = key;
    e->value = value;
  }
  return value;
}


/* does NOT check that symbol_id is valid */
char *symtab_name_for_id(symbol_table *st, int symbol_id)
{
  return st->keys[symbol_id];
}


/* EOF */
